//Function06.c Implementation File
//
// This program compares calling a function in a dynamically loaded 
// library to inline code.  Note that the library is loaded and the
// function is available OUTSIDE the iteration loop (the library is
// loaded once).  A reduced Lennard Jones pair energy calculation is
// used as a demo worker function.
//
// The library function is called from fnA()
// The inline code is in fnB()
//
// The code for opening and using the dynamic library was adapted from
// the Program Library HOW-TO of the Linux Documentation Project.  Note
// that there ARE some implmentation differences, such as if using HP-UX.

#include <sys/time.h>
#include <stdio.h>
#include <stdlib.h>
#include <dlfcn.h>

//defines
#define NumIts 10000000		//number of iterations

// computes a single particle pair energy via a call to a library function
void fnA()
{

	// general local declarations
	// integers
	int		counter;		//iteration counter
	
	// doubles
	double	r;				//inter particle distance
	double	sigma6;			//lj parameter, to sixth power
	double	Energy;			//returned energy
	double	(*LJEn)(double, double);	//pointer to library function

	void	*libhand;		//handle to library
	char	*error;			//error code for system calls

	// local declarations needed for timing
	// floats
	float ElapsedTime;	//ticks during loop
	
	// doubles
	double t1;		//ticks at start of loop
	double t2;		//ticks at loop end
	
	// structures needed for basic timing functions
	static struct timeval _tstart, _tstop;
	static struct timezone tz;
	
	
	// display entry message
	printf("\nMaking calls to dynamic library function...\n");
	
	// open the library
	libhand = dlopen ("./NumDemo/libNumDemo.so.1", RTLD_NOW);
    
	// check for open errors
	if (!libhand) {
		
		fputs (dlerror(), stderr);
        exit(1);
        
	}

	// make function available to current program
	LJEn = dlsym(libhand, "fnLJEnergy");

	//check for errors
	if ((error = dlerror()) != NULL) {

		fputs(error, stderr);
		exit(1);

	}

	// initialize r and sigma^6 with some demo values
	r = 1.234;		
	sigma6 = 0.5314;

	//get start time
	gettimeofday(&_tstart, &tz);
	
	for(counter=1; counter <= NumIts; ++counter){
    
		// call the function
		Energy = (*LJEn)(r, sigma6);

    }
	
	//stop the timer
	gettimeofday(&_tstop, &tz);

	//close the library
	dlclose(libhand);

	//get seconds from the structures
	t1 = (double)_tstart.tv_sec + (double)_tstart.tv_usec * 0.000001;
	t2 = (double)_tstop.tv_sec + (double)_tstop.tv_usec * 0.000001;
	
	//compute the elapsed time
	ElapsedTime = (float)(t2 - t1);

	//display the result
	printf("%s%g%s", "\nElapsed Time for library calls = ", ElapsedTime, " seconds\n\n");
	
}//end fnA()

// compute the reduced LJ Energy pair energy using inline code
void fnB()
{

	// general local declarations
	// integers
	int		counter;		//iteration counter
	
	// doubles
	double	r;				//inter particle distance
	double	r6;				//r^6
	double	r12;			//f^12
	double	sigma6;			//lj parameter, to sixth power
	double	sigma12;		//lj parameter, to 12th power
	double	Energy;			//returned energy

	// local declarations needed for timing
	// floats
	float ElapsedTime;	//ticks during loop
	
	// doubles
	double t1;		//ticks at start of loop
	double t2;		//ticks at loop end
	
	// structures needed for basic timing functions
	static struct timeval _tstart, _tstop;
	static struct timezone tz;
	
	
	// display entry message
	printf("\nComputing with inline code...\n");
	
	// initialize r and sigma^6 with some demo values
	r = 1.234;		
	sigma6 = 0.5314;

	//get start time
	gettimeofday(&_tstart, &tz);
	
	for(counter=1; counter <= NumIts; ++counter){
    
		//perform the computation directly
		//compute r^6
		r6 = r * r * r * r * r * r;
		
		//compute r^12
		r12 = r6 * r6;

		//compute sigma^6
		sigma12 = sigma6 * sigma6;

		//compute Energy
		Energy = sigma12/r12 - sigma6/r6;

    }
	
	//stop the timer
	gettimeofday(&_tstop, &tz);

	//get seconds from the structures
	t1 = (double)_tstart.tv_sec + (double)_tstart.tv_usec * 0.000001;
	t2 = (double)_tstop.tv_sec + (double)_tstop.tv_usec * 0.000001;
	
	//compute the elapsed time
	ElapsedTime = (float)(t2 - t1);

	//display the result
	printf("%s%g%s", "\nElapsed Time for inline code = ", ElapsedTime, " seconds\n\n");
	
}//end fnB()

//entry function
int main()
{
	//local declarations
	//integers
	int bTERM;		//menu flag

	//character strings
	char inchar;		//input character
	

	//initialize bTERM to NOT TERMINATE
	bTERM = 0;
	
	//clear terminal
	system("clear");
	
	do {
		
		//present menu to user
		printf("Enter the Number of the routine to run, and press <Enter> \n");
		printf("1\tCompute using dynamic library function calls\n");
		printf("2\tCompute using inline code\n");
		printf("q\tEnd this program\n");
	
		//get user input
		scanf("%s", &inchar);

		//inchar = getchar();

		//call the user selected function
		switch	(inchar){

			case '1': fnA();
				break;
	
			case '2': fnB();
				break;
		
			//select the terminate function	
			case 'q': printf("Ending This Program\n");
				
				//set flag to terminate
				bTERM = 1;
				break;

		}
		
	} while(!bTERM);

	return 0;

}
