// Loop03.c: implementation file
//
// This program demonstrates the unrolled loops for n x n matrix 
// multiplication.  However, using the unrolled loops necessitates
// the use of a decision structure when variable values for n are 
// possible in a routine, whereas the generic, rolled version does
// not require a decision since the dimension can be sent as a parameter.
//
// Therefore, in this demo, the program is performing 4x4 and 8x8
// matrix multiplications in random order using a variety of rolled
// algorithms compared to the unrolled one.
//
// A random number array and matrices are initialized in fnA().
//
// The rolled, generic routine matrix_nxn(int Dimen) is called with
// decision structures in fnB().
//
// The rolled, generic routine matrix_nxn(int Dimen) is called without
// decision structures in fnC().
//
// The rolled, generic routine matrix_nxn(int Dimen) is called with
// decision structures in fnD().
//
// The unrolled routine is called in fnE().

#include <sys/time.h>
#include <stdio.h>
#include <math.h>
#include <stdlib.h>

//defines
#define NumIts 500000		//number of iterations

//functions
void fnA();			
void fnB();
void fnC();
void fnD();
void fnE();

int matrix_nxn(int Dimen);
int matrix_4x4_rolled();
int matrix_8x8_rolled();
int matrix_4x4_unrolled();
int matrix_8x8_unrolled();
 			

// global declarations

//integers
int r[5000001];		//random 4's and 8's

// floats
float a[21][21];		//Matrix A
float b[21][21];		//Matrix B
float c[21][21];		//Result Matrix C

//initialize random number array and A and B matrices
void fnA()
{

	//local declarations
	//integers
	int rcounter;				//counter used to initialize random array
	int rowcounter;
	int colcounter;
	int randomval;


	//initialize the random number array
	for(rcounter = 1; rcounter <= NumIts; ++rcounter){

		//rand() returns an integer < about 32000
		randomval = rand();
		
		//want about half being 20, half being 4
		if(randomval < 16000) r[rcounter] = 4;
		else r[rcounter] = 8;
		
	}


	//initialize the A and B matrices to random values
	for(rowcounter=1; rowcounter <= 8; ++rowcounter){

		for(colcounter=1; colcounter <= 8; ++colcounter){

			a[rowcounter][colcounter] = (float)rand()/RAND_MAX;
			b[rowcounter][colcounter] = (float)rand()/RAND_MAX;

		}

	}	
	
}//end fnA()

// performs iterative generic nxn matrix multiplication with 
// decision structure 
void fnB()
{

	//general local declarations
	//integers
	int i;				//iteration counter
	int dummy;			//holder for return of function call
	
	//variables for time execution time
	//floats
	float ElapsedTime;	//ticks during loop
	
	//doubles
	double t1;		//ticks at start of loop
	double t2;		//ticks at loop end
	
	//structures needed for basic timing functions
	static struct timeval _tstart, _tstop;
	static struct timezone tz;

	//display entry message
	printf("\n nxn Matrix Multiplication with Decision Structure...\n");
	
	//get start time
	gettimeofday(&_tstart, &tz);
	
	for(i=1; i<= NumIts; ++i){
	
		//decision to determine which dimension to send
		if(r[i] == 8) dummy = matrix_nxn(8);
		else dummy = matrix_nxn(4);

	}
	
	//stop the timer
	gettimeofday(&_tstop, &tz);

	//get seconds from the structures
	t1 = (double)_tstart.tv_sec + (double)_tstart.tv_usec * 0.000001;
	t2 = (double)_tstop.tv_sec + (double)_tstop.tv_usec * 0.000001;
	
	//compute the elapsed time
	ElapsedTime = (float)(t2 - t1);

	//display the result
	printf("%s%g%s", "\nElapsed Time = ", ElapsedTime, " seconds\n");
	
}//end fnB()

// performs iterative generic nxn matrix multiplication without 
// decision structure
void fnC()
{

	//general local declarations
	//integers
	int i;				//iteration counter
	int dummy;			//holder for return of function call
	
	//variables for time execution time
	//floats
	float ElapsedTime;	//ticks during loop
	
	//doubles
	double t1;		//ticks at start of loop
	double t2;		//ticks at loop end
	
	//structures needed for basic timing functions
	static struct timeval _tstart, _tstop;
	static struct timezone tz;

	//display entry message
	printf("\n nxn Matrix Multiplication without Decision Structure...\n");
	
	//get start time
	gettimeofday(&_tstart, &tz);
	
	for(i=1; i<= NumIts; ++i){
	
		//no decision need be made, since r array holds the 
		//dimension itself
		dummy = matrix_nxn(r[i]);

	}
	
	//stop the timer
	gettimeofday(&_tstop, &tz);

	//get seconds from the structures
	t1 = (double)_tstart.tv_sec + (double)_tstart.tv_usec * 0.000001;
	t2 = (double)_tstop.tv_sec + (double)_tstop.tv_usec * 0.000001;
	
	//compute the elapsed time
	ElapsedTime = (float)(t2 - t1);

	//display the result
	printf("%s%g%s", "\nElapsed Time = ", ElapsedTime, " seconds\n");
	
}//end fnC()

// performs iterative explicit rolled matrix multiplication 
void fnD()
{

	//general local declarations
	//integers
	int i;				//iteration counter
	int dummy;			//holder for return of function call
	
	//variables for time execution time
	//floats
	float ElapsedTime;	//ticks during loop
	
	//doubles
	double t1;		//ticks at start of loop
	double t2;		//ticks at loop end
	
	//structures needed for basic timing functions
	static struct timeval _tstart, _tstop;
	static struct timezone tz;

	//display entry message
	printf("\n Explicit Rolled Matrix Multiplication ...\n");
	
	//get start time
	gettimeofday(&_tstart, &tz);
	
	for(i=1; i<= NumIts; ++i){
	
		//decision to determine which dimension to send
		if(r[i] == 8) dummy = matrix_8x8_rolled();
		else dummy = matrix_4x4_rolled();

	}
	
	//stop the timer
	gettimeofday(&_tstop, &tz);

	//get seconds from the structures
	t1 = (double)_tstart.tv_sec + (double)_tstart.tv_usec * 0.000001;
	t2 = (double)_tstop.tv_sec + (double)_tstop.tv_usec * 0.000001;
	
	//compute the elapsed time
	ElapsedTime = (float)(t2 - t1);

	//display the result
	printf("%s%g%s", "\nElapsed Time = ", ElapsedTime, " seconds\n");
	
}//end fnD()

// performs iterative unrolled matrix multiplication
void fnE()
{

	//general local declarations
	//integers
	int i;				//iteration counter
	int dummy;			//holder for return of function call
	
	//variables for time execution time
	//floats
	float ElapsedTime;	//ticks during loop
	
	//doubles
	double t1;		//ticks at start of loop
	double t2;		//ticks at loop end
	
	//structures needed for basic timing functions
	static struct timeval _tstart, _tstop;
	static struct timezone tz;

	//display entry message
	printf("\n Unrolled Matrix Multiplication ...\n");
	
	//get start time
	gettimeofday(&_tstart, &tz);
	
	for(i=1; i<= NumIts; ++i){
	
		//decision to determine which dimension to send
		if(r[i] == 8) dummy = matrix_8x8_unrolled();
		else dummy = matrix_4x4_unrolled();

	}
	
	//stop the timer
	gettimeofday(&_tstop, &tz);

	//get seconds from the structures
	t1 = (double)_tstart.tv_sec + (double)_tstart.tv_usec * 0.000001;
	t2 = (double)_tstop.tv_sec + (double)_tstop.tv_usec * 0.000001;
	
	//compute the elapsed time
	ElapsedTime = (float)(t2 - t1);

	//display the result
	printf("%s%g%s", "\nElapsed Time = ", ElapsedTime, " seconds\n");
	
}//end fnE()

//entry function
int main()
{
	//local declarations
	//integers
	int bTERM;		//menu flag

	//character strings
	char inchar;		//input character
	

	//initialize bTERM to NOT TERMINATE
	bTERM = 0;
	
	//clear terminal
	system("clear");

	//initialize random number array and matrices
	fnA();
	
	do {
		
		//present menu to user
		printf("Enter the Number of the routine to run, and press <Enter> \n");
		printf("1\tnxn Matrix Multiplication with Decision Structure\n");
		printf("2\tnxn Matrix Multiplication without Decision Structure\n");
		printf("3\tExplicit Dimension Rolled Matrix Multiplication\n");
		printf("4\tUnrolled Matrix Multiplication\n");
		printf("q\tEnd this program\n");
	
		//get user input
		scanf("%s", &inchar);

		//call the user selected function
		switch	(inchar){

			case '1': fnB();
				break;
	
			case '2': fnC();
				break;

			case '3': fnD();
				break;

			case '4': fnE();
				break;
		
			//select the terminate function	
			case 'q': printf("Ending This Program\n");
				
				//set flag to terminate
				bTERM = 1;
				break;

		}
		
	} while(!bTERM);

	return 0;

}

// Generic nxn matrix multiplication; dimension n is passed as integer
// the actual matrices are globals; actual function in application
// would better be coded more general with pointers.
int matrix_nxn(int Dimen)
{

	//local declarations
	int i;			//column counter
	int j;			//row counter
	int k;			//internal counter for matrix multiplication

	for(i=1; i<=Dimen; ++i){

		for(j=1; j<=Dimen; ++j){

			c[i][j] = 0;

			for(k=1; k<=Dimen; ++k){

				c[i][j] += a[i][k] * b[k][j];

			}

		}

	}

	return(0);

}


//Explicit 4x4 Rolled-Up matrix multiplication.  The dimension is explicit,
//so does not need to be passed.  The matrices are globals.  
int matrix_4x4_rolled()
{

	//local declarations
	int i;
	int j;
	int k;

	for(i=1; i <= 4; ++i){

		for(j=1; j <= 4; ++j){

			c[i][j] = 0;

			for(k=1; k <= 4; ++k){

				c[i][j] += a[i][k] * b[k][j];

			}

		}

	}

	return(0);

}

//Explicit 8x8 Rolled-Up matrix multiplication.  The dimension is explicit,
//so does not need to be passed.  The matrices are globals.  
int matrix_8x8_rolled()
{

	//local declarations
	int i;
	int j;
	int k;

	for(i=1; i <= 8; ++i){

		for(j=1; j <= 8; ++j){

			c[i][j] = 0;

			for(k=1; k <= 8; ++k){

				c[i][j] += a[i][k] * b[k][j];

			}

		}

	}

	return(0);

}

//Explicit, Unrolled 4x4 matrix multiplication
int matrix_4x4_unrolled()
{

	//j = 1
    c[1][ 1] = a[1][ 1] * b[1][ 1] + a[1][ 2] * b[2][ 1] + a[1][ 3] * b[3][ 1] + a[1][ 4] * b[4][ 1];
    c[1][ 2] = a[1][ 1] * b[1][ 2] + a[1][ 2] * b[2][ 2] + a[1][ 3] * b[3][ 2] + a[1][ 4] * b[4][ 2];
    c[1][ 3] = a[1][ 1] * b[1][ 3] + a[1][ 2] * b[2][ 3] + a[1][ 3] * b[3][ 3] + a[1][ 4] * b[4][ 3];
    c[1][ 4] = a[1][ 1] * b[1][ 4] + a[1][ 2] * b[2][ 4] + a[1][ 3] * b[3][ 4] + a[1][ 4] * b[4][ 4];
    
    //j = 2
    c[2][ 1] = a[2][ 1] * b[1][ 1] + a[2][ 2] * b[2][ 1] + a[2][ 3] * b[3][ 1] + a[2][ 4] * b[4][ 1];
    c[2][ 2] = a[2][ 1] * b[1][ 2] + a[2][ 2] * b[2][ 2] + a[2][ 3] * b[3][ 2] + a[2][ 4] * b[4][ 2];
    c[2][ 3] = a[2][ 1] * b[1][ 3] + a[2][ 2] * b[2][ 3] + a[2][ 3] * b[3][ 3] + a[2][ 4] * b[4][ 3];
    c[2][ 4] = a[2][ 1] * b[1][ 4] + a[2][ 2] * b[2][ 4] + a[2][ 3] * b[3][ 4] + a[2][ 4] * b[4][ 4];
    
    //j=3
    c[3][ 1] = a[3][ 1] * b[1][ 1] + a[3][ 2] * b[2][ 1] + a[3][ 3] * b[3][ 1] + a[3][ 4] * b[4][ 1];
    c[3][ 2] = a[3][ 1] * b[1][ 2] + a[3][ 2] * b[2][ 2] + a[3][ 3] * b[3][ 2] + a[3][ 4] * b[4][ 2];
    c[3][ 3] = a[3][ 1] * b[1][ 3] + a[3][ 2] * b[2][ 3] + a[3][ 3] * b[3][ 3] + a[3][ 4] * b[4][ 3];
    c[3][ 4] = a[3][ 1] * b[1][ 4] + a[3][ 2] * b[2][ 4] + a[3][ 3] * b[3][ 4] + a[3][ 4] * b[4][ 4];
    
    //j=4
    c[4][ 1] = a[4][ 1] * b[1][ 1] + a[4][ 2] * b[2][ 1] + a[4][ 3] * b[3][ 1] + a[4][ 4] * b[4][ 1];
    c[4][ 2] = a[4][ 1] * b[1][ 2] + a[4][ 2] * b[2][ 2] + a[4][ 3] * b[3][ 2] + a[4][ 4] * b[4][ 2];
    c[4][ 3] = a[4][ 1] * b[1][ 3] + a[4][ 2] * b[2][ 3] + a[4][ 3] * b[3][ 3] + a[4][ 4] * b[4][ 3];
    c[4][ 4] = a[4][ 1] * b[1][ 4] + a[4][ 2] * b[2][ 4] + a[4][ 3] * b[3][ 4] + a[4][ 4] * b[4][ 4];
    

	return(0);

}

//8x8 unrolled matrix multiplication
int matrix_8x8_unrolled()
{

	//i=1
    c[1][ 1] = a[1][1] * b[1][1] + a[1][2] * b[2][1] + a[1][3] * b[3][1] + a[1][4] * b[4][1]  
	 +	 a[1][ 5] * b[5][ 1] + a[1][ 6] * b[6][ 1] + a[1][ 7] * b[7][ 1] + a[1][ 8] * b[8][ 1];
    
    c[1][ 2] = a[1][ 1] * b[1][ 2] + a[1][ 2] * b[2][ 2] + a[1][ 3] * b[3][ 2] + a[1][ 4] * b[4][ 2]
     +   a[1][ 5] * b[5][ 2] + a[1][ 6] * b[6][ 2] + a[1][ 7] * b[7][ 2] + a[1][ 8] * b[8][ 2];
    
    c[1][ 3] = a[1][ 1] * b[1][ 3] + a[1][ 2] * b[2][ 3] + a[1][ 3] * b[3][ 3] + a[1][ 4] * b[4][ 3]
     +   a[1][ 5] * b[5][ 3] + a[1][ 6] * b[6][ 3] + a[1][ 7] * b[7][ 3] + a[1][ 8] * b[8][ 3];
    
    c[1][ 4] = a[1][ 1] * b[1][ 4] + a[1][ 2] * b[2][ 4] + a[1][ 3] * b[3][ 4] + a[1][ 4] * b[4][ 4]
     +  a[1][ 4] * b[5][ 4] + a[1][ 6] * b[6][ 4] + a[1][ 7] * b[7][ 4] + a[1][ 8] * b[8][ 4];
    
    c[1][ 5] = a[1][ 1] * b[1][ 5] + a[1][ 2] * b[2][ 5] + a[1][ 3] * b[3][ 5] + a[1][ 4] * b[4][ 5]
     +   a[1][ 5] * b[5][ 5] + a[1][ 6] * b[6][ 5] + a[1][ 7] * b[7][ 5] + a[1][ 8] * b[8][ 5];
    
    c[1][ 6] = a[1][ 1] * b[1][ 6] + a[1][ 2] * b[2][ 6] + a[1][ 3] * b[3][ 6] + a[1][ 4] * b[4][ 6]
     +   a[1][ 5] * b[5][ 6] + a[1][ 6] * b[6][ 6] + a[1][ 7] * b[7][ 6] + a[1][ 8] * b[8][ 6];
    
    c[1][ 7] = a[1][ 1] * b[1][ 7] + a[1][ 2] * b[2][ 7] + a[1][ 3] * b[3][ 7] + a[1][ 4] * b[4][ 7]
     +   a[1][ 5] * b[5][ 7] + a[1][ 6] * b[6][ 7] + a[1][ 7] * b[7][ 7] + a[1][ 8] * b[8][ 7];
    
    c[1][ 8] = a[1][ 1] * b[1][ 8] + a[1][ 2] * b[2][ 8] + a[1][ 3] * b[3][ 8] + a[1][ 4] * b[4][ 8]
     +   a[1][ 5] * b[5][ 8] + a[1][ 6] * b[6][ 8] + a[1][ 7] * b[7][ 8] + a[1][ 8] * b[8][ 8];
    
    //i=2
    
    c[2][ 1] = a[2][ 1] * b[1][ 1] + a[2][ 2] * b[2][ 1] + a[2][ 3] * b[3][ 1] + a[2][ 4] * b[4][ 1]
     +   a[2][ 5] * b[5][ 1] + a[2][ 6] * b[6][ 1] + a[2][ 7] * b[7][ 1] + a[2][ 8] * b[8][ 1];
    
    c[2][ 2] = a[2][ 1] * b[1][ 2] + a[2][ 2] * b[2][ 2] + a[2][ 3] * b[3][ 2] + a[2][ 4] * b[4][ 2]
     +   a[2][ 5] * b[5][ 2] + a[2][ 6] * b[6][ 2] + a[2][ 7] * b[7][ 2] + a[2][ 8] * b[8][ 2];
    
    c[2][ 3] = a[2][ 1] * b[1][ 3] + a[2][ 2] * b[2][ 3] + a[2][ 3] * b[3][ 3] + a[2][ 4] * b[4][ 3]
     +   a[2][ 5] * b[5][ 3] + a[2][ 6] * b[6][ 3] + a[2][ 7] * b[7][ 3] + a[2][ 8] * b[8][ 3];
    
    c[2][ 4] = a[2][ 1] * b[1][ 4] + a[2][ 2] * b[2][ 4] + a[2][ 3] * b[3][ 4] + a[2][ 4] * b[4][ 4]
     +   a[2][ 4] * b[5][ 4] + a[2][ 6] * b[6][ 4] + a[2][ 7] * b[7][ 4] + a[2][ 8] * b[8][ 4];
    
    c[2][ 5] = a[2][ 1] * b[1][ 5] + a[2][ 2] * b[2][ 5] + a[2][ 3] * b[3][ 5] + a[2][ 4] * b[4][ 5]
     +   a[2][ 5] * b[5][ 5] + a[2][ 6] * b[6][ 5] + a[2][ 7] * b[7][ 5] + a[2][ 8] * b[8][ 5];
    
    c[2][ 6] = a[2][ 1] * b[1][ 6] + a[2][ 2] * b[2][ 6] + a[2][ 3] * b[3][ 6] + a[2][ 4] * b[4][ 6]
     +   a[2][ 5] * b[5][ 6] + a[2][ 6] * b[6][ 6] + a[2][ 7] * b[7][ 6] + a[2][ 8] * b[8][ 6];
    
    c[2][ 7] = a[2][ 1] * b[1][ 7] + a[2][ 2] * b[2][ 7] + a[2][ 3] * b[3][ 7] + a[2][ 4] * b[4][ 7]
     +   a[2][ 5] * b[5][ 7] + a[2][ 6] * b[6][ 7] + a[2][ 7] * b[7][ 7] + a[2][ 8] * b[8][ 7];
    
    c[2][ 8] = a[2][ 1] * b[1][ 8] + a[2][ 2] * b[2][ 8] + a[2][ 3] * b[3][ 8] + a[2][ 4] * b[4][ 8]
     +   a[2][ 5] * b[5][ 8] + a[2][ 6] * b[6][ 8] + a[2][ 7] * b[7][ 8] + a[2][ 8] * b[8][ 8];

    //i=3
    
    c[3][ 1] = a[3][ 1] * b[1][ 1] + a[3][ 2] * b[2][ 1] + a[3][ 3] * b[3][ 1] + a[3][ 4] * b[4][ 1]
     +   a[3][ 5] * b[5][ 1] + a[3][ 6] * b[6][ 1] + a[3][ 7] * b[7][ 1] + a[3][ 8] * b[8][ 1];

    c[3][ 2] = a[3][ 1] * b[1][ 2] + a[3][ 2] * b[2][ 2] + a[3][ 3] * b[3][ 2] + a[3][ 4] * b[4][ 2]
     +   a[3][ 5] * b[5][ 2] + a[3][ 6] * b[6][ 2] + a[3][ 7] * b[7][ 2] + a[3][ 8] * b[8][ 2];
    
    c[3][ 3] = a[3][ 1] * b[1][ 3] + a[3][ 2] * b[2][ 3] + a[3][ 3] * b[3][ 3] + a[3][ 4] * b[4][ 3]
     +   a[3][ 5] * b[5][ 3] + a[3][ 6] * b[6][ 3] + a[3][ 7] * b[7][ 3] + a[3][ 8] * b[8][ 3];
    
    c[3][ 4] = a[3][ 1] * b[1][ 4] + a[3][ 2] * b[2][ 4] + a[3][ 3] * b[3][ 4] + a[3][ 4] * b[4][ 4]
     +   a[3][ 4] * b[5][ 4] + a[3][ 6] * b[6][ 4] + a[3][ 7] * b[7][ 4] + a[3][ 8] * b[8][ 4];
    
    c[3][ 5] = a[3][ 1] * b[1][ 5] + a[3][ 2] * b[2][ 5] + a[3][ 3] * b[3][ 5] + a[3][ 4] * b[4][ 5]
     +   a[3][ 5] * b[5][ 5] + a[3][ 6] * b[6][ 5] + a[3][ 7] * b[7][ 5] + a[3][ 8] * b[8][ 5];
    
    c[3][ 6] = a[3][ 1] * b[1][ 6] + a[3][ 2] * b[2][ 6] + a[3][ 3] * b[3][ 6] + a[3][ 4] * b[4][ 6]
     +   a[3][ 5] * b[5][ 6] + a[3][ 6] * b[6][ 6] + a[3][ 7] * b[7][ 6] + a[3][ 8] * b[8][ 6];
    
    c[3][ 7] = a[3][ 1] * b[1][ 7] + a[3][ 2] * b[2][ 7] + a[3][ 3] * b[3][ 7] + a[3][ 4] * b[4][ 7]
     +   a[3][ 5] * b[5][ 7] + a[3][ 6] * b[6][ 7] + a[3][ 7] * b[7][ 7] + a[3][ 8] * b[8][ 7];
    
    c[3][ 8] = a[3][ 1] * b[1][ 8] + a[3][ 2] * b[2][ 8] + a[3][ 3] * b[3][ 8] + a[3][ 4] * b[4][ 8]
     +   a[3][ 5] * b[5][ 8] + a[3][ 6] * b[6][ 8] + a[3][ 7] * b[7][ 8] + a[3][ 8] * b[8][ 8];

    //i=4
    
    c[4][ 1] = a[4][ 1] * b[1][ 1] + a[4][ 2] * b[2][ 1] + a[4][ 3] * b[3][ 1] + a[4][ 4] * b[4][ 1]
     +   a[4][ 5] * b[5][ 1] + a[4][ 6] * b[6][ 1] + a[4][ 7] * b[7][ 1] + a[4][ 8] * b[8][ 1];
    
    c[4][ 2] = a[4][ 1] * b[1][ 2] + a[4][ 2] * b[2][ 2] + a[4][ 3] * b[3][ 2] + a[4][ 4] * b[4][ 2]
     +   a[4][ 5] * b[5][ 2] + a[4][ 6] * b[6][ 2] + a[4][ 7] * b[7][ 2] + a[4][ 8] * b[8][ 2];
    
    c[4][ 3] = a[4][ 1] * b[1][ 3] + a[4][ 2] * b[2][ 3] + a[4][ 3] * b[3][ 3] + a[4][ 4] * b[4][ 3]
     +   a[4][ 5] * b[5][ 3] + a[4][ 6] * b[6][ 3] + a[4][ 7] * b[7][ 3] + a[4][ 8] * b[8][ 3];
    
    c[4][ 4] = a[4][ 1] * b[1][ 4] + a[4][ 2] * b[2][ 4] + a[4][ 3] * b[3][ 4] + a[4][ 4] * b[4][ 4]
     +   a[4][ 4] * b[5][ 4] + a[4][ 6] * b[6][ 4] + a[4][ 7] * b[7][ 4] + a[4][ 8] * b[8][ 4];
    
    c[4][ 5] = a[4][ 1] * b[1][ 5] + a[4][ 2] * b[2][ 5] + a[4][ 3] * b[3][ 5] + a[4][ 4] * b[4][ 5]
     +   a[4][ 5] * b[5][ 5] + a[4][ 6] * b[6][ 5] + a[4][ 7] * b[7][ 5] + a[4][ 8] * b[8][ 5];
    
    c[4][ 6] = a[4][ 1] * b[1][ 6] + a[4][ 2] * b[2][ 6] + a[4][ 3] * b[3][ 6] + a[4][ 4] * b[4][ 6]
     +   a[4][ 5] * b[5][ 6] + a[4][ 6] * b[6][ 6] + a[4][ 7] * b[7][ 6] + a[4][ 8] * b[8][ 6];
    
    c[4][ 7] = a[4][ 1] * b[1][ 7] + a[4][ 2] * b[2][ 7] + a[4][ 3] * b[3][ 7] + a[4][ 4] * b[4][ 7]
     +   a[4][ 5] * b[5][ 7] + a[4][ 6] * b[6][ 7] + a[4][ 7] * b[7][ 7] + a[4][ 8] * b[8][ 7];
    
    c[4][ 8] = a[4][ 1] * b[1][ 8] + a[4][ 2] * b[2][ 8] + a[4][ 3] * b[3][ 8] + a[4][ 4] * b[4][ 8]
     +   a[4][ 5] * b[5][ 8] + a[4][ 6] * b[6][ 8] + a[4][ 7] * b[7][ 8] + a[4][ 8] * b[8][ 8];
    
    //i=5
    
    c[5][ 1] = a[5][ 1] * b[1][ 1] + a[5][ 2] * b[2][ 1] + a[5][ 3] * b[3][ 1] + a[5][ 4] * b[4][ 1]
     +   a[5][ 5] * b[5][ 1] + a[5][ 6] * b[6][ 1] + a[5][ 7] * b[7][ 1] + a[5][ 8] * b[8][ 1];
    
    c[5][ 2] = a[5][ 1] * b[1][ 2] + a[5][ 2] * b[2][ 2] + a[5][ 3] * b[3][ 2] + a[5][ 4] * b[4][ 2]
     +   a[5][ 5] * b[5][ 2] + a[5][ 6] * b[6][ 2] + a[5][ 7] * b[7][ 2] + a[5][ 8] * b[8][ 2];
    
    c[5][ 3] = a[5][ 1] * b[1][ 3] + a[5][ 2] * b[2][ 3] + a[5][ 3] * b[3][ 3] + a[5][ 4] * b[4][ 3]
     +   a[5][ 5] * b[5][ 3] + a[5][ 6] * b[6][ 3] + a[5][ 7] * b[7][ 3] + a[5][ 8] * b[8][ 3];
    
    c[5][ 4] = a[5][ 1] * b[1][ 4] + a[5][ 2] * b[2][ 4] + a[5][ 3] * b[3][ 4] + a[5][ 4] * b[4][ 4]
     +   a[5][ 4] * b[5][ 4] + a[5][ 6] * b[6][ 4] + a[5][ 7] * b[7][ 4] + a[5][ 8] * b[8][ 4];
    
    c[5][ 5] = a[5][ 1] * b[1][ 5] + a[5][ 2] * b[2][ 5] + a[5][ 3] * b[3][ 5] + a[5][ 4] * b[4][ 5]
     +   a[5][ 5] * b[5][ 5] + a[5][ 6] * b[6][ 5] + a[5][ 7] * b[7][ 5] + a[5][ 8] * b[8][ 5];
    
    c[5][ 6] = a[5][ 1] * b[1][ 6] + a[5][ 2] * b[2][ 6] + a[5][ 3] * b[3][ 6] + a[5][ 4] * b[4][ 6]
     +   a[5][ 5] * b[5][ 6] + a[5][ 6] * b[6][ 6] + a[5][ 7] * b[7][ 6] + a[5][ 8] * b[8][ 6];
    
    c[5][ 7] = a[5][ 1] * b[1][ 7] + a[5][ 2] * b[2][ 7] + a[5][ 3] * b[3][ 7] + a[5][ 4] * b[4][ 7]
     +   a[5][ 5] * b[5][ 7] + a[5][ 6] * b[6][ 7] + a[5][ 7] * b[7][ 7] + a[5][ 8] * b[8][ 7];
    
    c[5][ 8] = a[5][ 1] * b[1][ 8] + a[5][ 2] * b[2][ 8] + a[5][ 3] * b[3][ 8] + a[5][ 4] * b[4][ 8]
     +   a[5][ 5] * b[5][ 8] + a[5][ 6] * b[6][ 8] + a[5][ 7] * b[7][ 8] + a[5][ 8] * b[8][ 8];

    //i=6
    
    c[6][ 1] = a[6][ 1] * b[1][ 1] + a[6][ 2] * b[2][ 1] + a[6][ 3] * b[3][ 1] + a[6][ 4] * b[4][ 1]
     +   a[6][ 5] * b[5][ 1] + a[6][ 6] * b[6][ 1] + a[6][ 7] * b[7][ 1] + a[6][ 8] * b[8][ 1];
    
    c[6][ 2] = a[6][ 1] * b[1][ 2] + a[6][ 2] * b[2][ 2] + a[6][ 3] * b[3][ 2] + a[6][ 4] * b[4][ 2]
     +   a[6][ 5] * b[5][ 2] + a[6][ 6] * b[6][ 2] + a[6][ 7] * b[7][ 2] + a[6][ 8] * b[8][ 2];
    
    c[6][ 3] = a[6][ 1] * b[1][ 3] + a[6][ 2] * b[2][ 3] + a[6][ 3] * b[3][ 3] + a[6][ 4] * b[4][ 3]
     +   a[6][ 5] * b[5][ 3] + a[6][ 6] * b[6][ 3] + a[6][ 7] * b[7][ 3] + a[6][ 8] * b[8][ 3];
    
    c[6][ 4] = a[6][ 1] * b[1][ 4] + a[6][ 2] * b[2][ 4] + a[6][ 3] * b[3][ 4] + a[6][ 4] * b[4][ 4]
     +   a[6][ 4] * b[5][ 4] + a[6][ 6] * b[6][ 4] + a[6][ 7] * b[7][ 4] + a[6][ 8] * b[8][ 4];
    
    c[6][ 5] = a[6][ 1] * b[1][ 5] + a[6][ 2] * b[2][ 5] + a[6][ 3] * b[3][ 5] + a[6][ 4] * b[4][ 5]
     +   a[6][ 5] * b[5][ 5] + a[6][ 6] * b[6][ 5] + a[6][ 7] * b[7][ 5] + a[6][ 8] * b[8][ 5];
    
    c[6][ 6] = a[6][ 1] * b[1][ 6] + a[6][ 2] * b[2][ 6] + a[6][ 3] * b[3][ 6] + a[6][ 4] * b[4][ 6]
     +   a[6][ 5] * b[5][ 6] + a[6][ 6] * b[6][ 6] + a[6][ 7] * b[7][ 6] + a[6][ 8] * b[8][ 6];
    
    c[6][ 7] = a[6][ 1] * b[1][ 7] + a[6][ 2] * b[2][ 7] + a[6][ 3] * b[3][ 7] + a[6][ 4] * b[4][ 7]
     +   a[6][ 5] * b[5][ 7] + a[6][ 6] * b[6][ 7] + a[6][ 7] * b[7][ 7] + a[6][ 8] * b[8][ 7];
    
    c[6][ 8] = a[6][ 1] * b[1][ 8] + a[6][ 2] * b[2][ 8] + a[6][ 3] * b[3][ 8] + a[6][ 4] * b[4][ 8]
     +   a[6][ 5] * b[5][ 8] + a[6][ 6] * b[6][ 8] + a[6][ 7] * b[7][ 8] + a[6][ 8] * b[8][ 8];

    //i=7
    
    c[7][ 1] = a[7][ 1] * b[1][ 1] + a[7][ 2] * b[2][ 1] + a[7][ 3] * b[3][ 1] + a[7][ 4] * b[4][ 1]
     +   a[7][ 5] * b[5][ 1] + a[7][ 6] * b[6][ 1] + a[7][ 7] * b[7][ 1] + a[7][ 8] * b[8][ 1];
    
    c[7][ 2] = a[7][ 1] * b[1][ 2] + a[7][ 2] * b[2][ 2] + a[7][ 3] * b[3][ 2] + a[7][ 4] * b[4][ 2]
     +   a[7][ 5] * b[5][ 2] + a[7][ 6] * b[6][ 2] + a[7][ 7] * b[7][ 2] + a[7][ 8] * b[8][ 2];
    
    c[7][ 3] = a[7][ 1] * b[1][ 3] + a[7][ 2] * b[2][ 3] + a[7][ 3] * b[3][ 3] + a[7][ 4] * b[4][ 3]
     +   a[7][ 5] * b[5][ 3] + a[7][ 6] * b[6][ 3] + a[7][ 7] * b[7][ 3] + a[7][ 8] * b[8][ 3];
    
    c[7][ 4] = a[7][ 1] * b[1][ 4] + a[7][ 2] * b[2][ 4] + a[7][ 3] * b[3][ 4] + a[7][ 4] * b[4][ 4]
     +   a[7][ 4] * b[5][ 4] + a[7][ 6] * b[6][ 4] + a[7][ 7] * b[7][ 4] + a[7][ 8] * b[8][ 4];
    
    c[7][ 5] = a[7][ 1] * b[1][ 5] + a[7][ 2] * b[2][ 5] + a[7][ 3] * b[3][ 5] + a[7][ 4] * b[4][ 5]
     +   a[7][ 5] * b[5][ 5] + a[7][ 6] * b[6][ 5] + a[7][ 7] * b[7][ 5] + a[7][ 8] * b[8][ 5];
    
    c[7][ 6] = a[7][ 1] * b[1][ 6] + a[7][ 2] * b[2][ 6] + a[7][ 3] * b[3][ 6] + a[7][ 4] * b[4][ 6]
     +   a[7][ 5] * b[5][ 6] + a[7][ 6] * b[6][ 6] + a[7][ 7] * b[7][ 6] + a[7][ 8] * b[8][ 6];
    
    c[7][ 7] = a[7][ 1] * b[1][ 7] + a[7][ 2] * b[2][ 7] + a[7][ 3] * b[3][ 7] + a[7][ 4] * b[4][ 7]
     +   a[7][ 5] * b[5][ 7] + a[7][ 6] * b[6][ 7] + a[7][ 7] * b[7][ 7] + a[7][ 8] * b[8][ 7];
    
    c[7][ 8] = a[7][ 1] * b[1][ 8] + a[7][ 2] * b[2][ 8] + a[7][ 3] * b[3][ 8] + a[7][ 4] * b[4][ 8]
     +   a[7][ 5] * b[5][ 8] + a[7][ 6] * b[6][ 8] + a[7][ 7] * b[7][ 8] + a[7][ 8] * b[8][ 8];

    //i=8
    
    c[8][ 1] = a[8][ 1] * b[1][ 1] + a[8][ 2] * b[2][ 1] + a[8][ 3] * b[3][ 1] + a[8][ 4] * b[4][ 1]
     +   a[8][ 5] * b[5][ 1] + a[8][ 6] * b[6][ 1] + a[8][ 7] * b[7][ 1] + a[8][ 8] * b[8][ 1];
    
    c[8][ 2] = a[8][ 1] * b[1][ 2] + a[8][ 2] * b[2][ 2] + a[8][ 3] * b[3][ 2] + a[8][ 4] * b[4][ 2]
     +   a[8][ 5] * b[5][ 2] + a[8][ 6] * b[6][ 2] + a[8][ 7] * b[7][ 2] + a[8][ 8] * b[8][ 2];
    
    c[8][ 3] = a[8][ 1] * b[1][ 3] + a[8][ 2] * b[2][ 3] + a[8][ 3] * b[3][ 3] + a[8][ 4] * b[4][ 3]
     +   a[8][ 5] * b[5][ 3] + a[8][ 6] * b[6][ 3] + a[8][ 7] * b[7][ 3] + a[8][ 8] * b[8][ 3];
    
    c[8][ 4] = a[8][ 1] * b[1][ 4] + a[8][ 2] * b[2][ 4] + a[8][ 3] * b[3][ 4] + a[8][ 4] * b[4][ 4]
     +   a[8][ 4] * b[5][ 4] + a[8][ 6] * b[6][ 4] + a[8][ 7] * b[7][ 4] + a[8][ 8] * b[8][ 4];
    
    c[8][ 5] = a[8][ 1] * b[1][ 5] + a[8][ 2] * b[2][ 5] + a[8][ 3] * b[3][ 5] + a[8][ 4] * b[4][ 5]
     +   a[8][ 5] * b[5][ 5] + a[8][ 6] * b[6][ 5] + a[8][ 7] * b[7][ 5] + a[8][ 8] * b[8][ 5];
    
    c[8][ 6] = a[8][ 1] * b[1][ 6] + a[8][ 2] * b[2][ 6] + a[8][ 3] * b[3][ 6] + a[8][ 4] * b[4][ 6]
     +   a[8][ 5] * b[5][ 6] + a[8][ 6] * b[6][ 6] + a[8][ 7] * b[7][ 6] + a[8][ 8] * b[8][ 6];
    
    c[8][ 7] = a[8][ 1] * b[1][ 7] + a[8][ 2] * b[2][ 7] + a[8][ 3] * b[3][ 7] + a[8][ 4] * b[4][ 7]
     +   a[8][ 5] * b[5][ 7] + a[8][ 6] * b[6][ 7] + a[8][ 7] * b[7][ 7] + a[8][ 8] * b[8][ 7];
    
    c[8][ 8] = a[8][ 1] * b[1][ 8] + a[8][ 2] * b[2][ 8] + a[8][ 3] * b[3][ 8] + a[8][ 4] * b[4][ 8]
     +   a[8][ 5] * b[5][ 8] + a[8][ 6] * b[6][ 8] + a[8][ 7] * b[7][ 8] + a[8][ 8] * b[8][ 8];

    return(0);

}
