///////////////////// Vector Implementation /////////////////////////
//                                                                 //
//                   x87 based Vector Routines                     //
//                                                                 //
//  Written by:  John S. Riley                                     //
//                                                                 //
//  This file contains a variety of vector manipulation routines   //
//  written for generic x87 floating-point processors.             //
//                                                                 //
//  The routines use no explicit SIMD instructions, but may be     //
//  compiled using compiler generated SSE or 3dNow instructions.   //
//                                                                 //
//  The vector routines are explicit dimension, using unrolled     //
//  loops. Vector data are passed to the functions as pointers.    //
//	Calls to these routines will be faster than calls to the	   //
//	generic routines for two reasons:							   //
//																   //
//		(1)	the unrolled algorithms have no loop overhead or	   //
//			index computation inside the loops					   //
//																   //
//		(2) fewer parameters passed in the function call,		   //
//			specifically, no dimensions need be passed.			   //
//                                                                 //
/////////////////////////////////////////////////////////////////////

///////////////////////// Functions in this File ////////////////////
//                                                                 //
//	VectorDotProd2d:	Dot product of two 2d vectors			   //
//	VectorDotProd3d:	Dot product of two 3d vectors			   //
//	VectorAdd2d:		Addition of two 2d vectors				   //
//	VectorAdd3d:		Addition of two 3d vectors				   //
//	VectorNegate2d:		Given A, computes -A, for 2d A			   //
//	VectorNegate3d:		Given A, computes -A, for 3d A			   //
//	VectorAngle2d:		Computes the angle between two 2d vectors  //
//	VectorAngle3d:		Angle between two 3d vectors			   //
//	VectorCrossLen2d:	Length of the cross product, 2d vectors	   //
//	VectorCrossLen3d:	Length of the cross product, 3d vectors	   //
//	VectorScale2d:		Multiply a 2d vector by a scalar		   //
//	VectorScale3d:		Multiply a 3d vector by a scalar		   //
//	VectorLength2d:		Length of 2d vector						   //
//	VectorLength3d:		Length of 3d vector						   //   
//																   //
/////////////////////////////////////////////////////////////////////

///////////////////////// INCLUDES //////////////////////////////////

#include "Vector.h"
#include "stdlib.h"
#include "math.h"

///////////////////////// FUNCTIONS /////////////////////////////////

//////////////////////// VectorDotProd2d ////////////////////////////
//																   //
//	This function computes the dot product of two 2d	 		   //
//  vectors														   //
//																   //
//	C = A . B, computed as C = a[1]b[1] + a[2]b[2]				   //
//																   //
//	INPUT: Passed as parameters:								   //
//																   //
//			*VectorA	float	Pointer to A data				   //
//			*VectorB	float	Pointer to B data				   //
//			*DotProd	float	Pointer to result, C			   //
//																   //
//	OUTPUT:	Return value is errno, math error code defined in	   //
//			C standard library.									   //
//																   //
//			Returns 0 if no math error.							   //
//																   //
/////////////////////////////////////////////////////////////////////

int VectorDotProd2d(
					float	*VectorA,
					float	*VectorB,
					float	*DotProd)
{

	*DotProd = *VectorA * *VectorB + *(VectorA + 1) * *(VectorB + 1);

	return(errno);

} // End VectorDotProd2d

//////////////////////// VectorDotProd3d ////////////////////////////
//																   //
//	This function computes the dot product of two 3d	 		   //
//  vectors														   //
//																   //
//	C = A . B, computed as C = a[1]b[1] + a[2]b[2] + a[3]b[3]				   //
//																   //
//	INPUT: Passed as parameters:								   //
//																   //
//			*VectorA	float	Pointer to A data				   //
//			*VectorB	float	Pointer to B data				   //
//			*DotProd	float	Pointer to result, C			   //
//																   //
//	OUTPUT:	Return value is errno, math error code defined in	   //
//			C standard library.									   //
//																   //
//			Returns 0 if no math error.							   //
//																   //
/////////////////////////////////////////////////////////////////////

int VectorDotProd3d(
					float	*VectorA,
					float	*VectorB,
					float	*DotProd)

{

	*DotProd = (*VectorA * *VectorB) + (*(VectorA + 1) * *(VectorB + 1))
			+ (*(VectorA + 2) * *(VectorB + 2));

	return(errno);

} // End VectorDotProd3d

//////////////////////// VectorAdd2d ////////////////////////////////
//																   //
//	This function adds  two 2d vectors			 		 		   //
//  															   //
//																   //
//	C = A + B, computed as  c[1] = a[1] + b[1]	 				   //
//							c[2] = a[2] + b[2]					   //
//																   //
//	INPUT: Passed as parameters:								   //
//																   //
//			*VectorA	float	Pointer to A data				   //
//			*VectorB	float	Pointer to B data				   //
//			*Sum		float	Pointer to result, C			   //
//																   //
//	OUTPUT:	Return value is errno, math error code defined in	   //
//			C standard library.									   //
//																   //
//			Returns 0 if no math error.							   //
//																   //
/////////////////////////////////////////////////////////////////////

int VectorAdd2d(
				float	*VectorA,
				float	*VectorB,
				float	*Sum)
{

	*Sum = *VectorA + *VectorB;
	*(Sum + 1) = *(VectorA + 1) + *(VectorB + 1);

	return(errno);

} // End VectorAdd2d

//////////////////////// VectorAdd3d ////////////////////////////////
//																   //
//	This function adds two 3d vectors.					 		   //
//  															   //
//																   //
//	C = A + B, computed as  c[1] = a[1] + b[1]	 				   //
//							c[2] = a[2] + b[2]					   //
//							c[3] = a[3] + b[3]					   //
//																   //
//	INPUT: Passed as parameters:								   //
//																   //
//			*VectorA	float	Pointer to A data				   //
//			*VectorB	float	Pointer to B data				   //
//			*Sum		float	Pointer to result, C			   //
//																   //
//	OUTPUT:	Return value is errno, math error code defined in	   //
//			C standard library.									   //
//																   //
//			Returns 0 if no math error.							   //
//																   //
/////////////////////////////////////////////////////////////////////

int VectorAdd3d(
				float	*VectorA,
				float	*VectorB,
				float	*Sum)
{

	*Sum = *VectorA + *VectorB;
	*(Sum + 1) = *(VectorA + 1) + *(VectorB + 1);
	*(Sum + 2) = *(VectorA + 2) + *(VectorB + 2);

	return(errno);

} // End VectorAdd3d

//////////////////////// VectorNegate2d ////////////////////////////////
//																   //
//	This function negates the a 2d vector	    		 		   //
//																   //
//																   //
//	B = -A														   //
//																   //
//	INPUT: Passed as parameters:								   //
//																   //
//			*VectorA	float	Pointer to A data				   //
//			*VectorB	float	Pointer to B data, -A			   //
//																   //
//	OUTPUT:	Return value is errno, math error code defined in	   //
//			C standard library.									   //
//																   //
//			Returns 0 if no math error.							   //
//																   //
/////////////////////////////////////////////////////////////////////

int VectorNegate2d(
				float	*VectorA,
				float	*VectorB)
{

	*VectorB = -1 * *VectorA;
	*(VectorB + 1) = -1 * *(VectorA + 1);

	return(errno);

} // End VectorNegate2d

//////////////////////// VectorNegate3d ////////////////////////////////
//																   //
//	This function negates the a 3d vector	    		 		   //
//																   //
//																   //
//	B = -A														   //
//																   //
//	INPUT: Passed as parameters:								   //
//																   //
//			*VectorA	float	Pointer to A data				   //
//			*VectorB	float	Pointer to B data, -A			   //
//																   //
//	OUTPUT:	Return value is errno, math error code defined in	   //
//			C standard library.									   //
//																   //
//			Returns 0 if no math error.							   //
//																   //
/////////////////////////////////////////////////////////////////////

int VectorNegate3d(
				float	*VectorA,
				float	*VectorB)
{

	*VectorB = -1 * *VectorA;
	*(VectorB + 1) = -1 * *(VectorA + 1);
	*(VectorB + 2) = -1 * *(VectorA + 2);

	return(errno);

} // End VectorNegate3d

//////////////////////// VectorScale2d //////////////////////////////
//																   //
//	This function multiplies a 2d vector by a scalar	 		   //
//																   //
//	B = cA, where b[i] = c * a[i]								   //
//																   //
//	INPUT: Passed as parameters:								   //
//																   //
//		*VectorA		float	Pointer to A data				   //
//		Scale			float	Scalar							   //
//		*ScaledVector:	float	Pointer to scaled vector array	   //
//																   //
//	OUTPUT:	Return value is errno, math error code defined in	   //
//			C standard library.									   //
//																   //
//			Returns 0 if no math error.							   //
//																   //
/////////////////////////////////////////////////////////////////////

int	VectorScale2d(
				  float	*VectorA,
				  float	Scale,
				  float	*ScaledVector)
{

	*ScaledVector = *VectorA * Scale;
	*(ScaledVector + 1) = *(VectorA + 1) * Scale;

	return(errno);

} // End VectorScale2d

//////////////////////// VectorScale3d //////////////////////////////
//																   //
//	This function multiplies a 3d vector by a scalar	 		   //
//																   //
//	B = cA, where b[i] = c * a[i]								   //
//																   //
//	INPUT: Passed as parameters:								   //
//																   //
//		*VectorA		float	Pointer to A data				   //
//		Scale			float	Scalar							   //
//		*ScaledVector:	float	Pointer to scaled vector array	   //
//																   //
//	OUTPUT:	Return value is errno, math error code defined in	   //
//			C standard library.									   //
//																   //
//			Returns 0 if no math error.							   //
//																   //
/////////////////////////////////////////////////////////////////////

int	VectorScale3d(
				  float	*VectorA,
				  float	Scale,
				  float	*ScaledVector)
{

	*ScaledVector = *VectorA * Scale;
	*(ScaledVector + 1) = *(VectorA + 1) * Scale;
	*(ScaledVector + 2) = *(VectorA + 2) * Scale;

	return(errno);

} // End VectorScale3d

//////////////////////// VectorLength2d /////////////////////////////
//																   //
//	This function computes the length of a 2d vector, assuming one //
//	one end is at the origin.									   //
//																   //
//	Length = sqrt(a[0]^2 + a[1]^2)								   //
//																   //
//	INPUT: Passed as parameters:								   //
//																   //
//		*Vector			float	Pointer to vector data			   //
//		*Len:			float	Pointer to store length			   //
//																   //
//	OUTPUT:	Return value is errno, math error code defined in	   //
//			C standard library.									   //
//																   //
//			Returns 0 if no math error.							   //
//																   //
/////////////////////////////////////////////////////////////////////

int VectorLength2d(
				   float	*Vector,
				   float	*Len)
{
	//local declarations
	float	t1;
	float	t2;

	t1 = *Vector * *Vector;
	t2 = *(Vector + 1) * *(Vector + 1);

	*Len = (float)sqrt(t1 + t2);

	return(0);

} // End VectorLength2d

//////////////////////// VectorLength3d /////////////////////////////
//																   //
//	This function computes the length of a 3d vector, assuming one //
//	one end is at the origin.									   //
//																   //
//	Length = sqrt(a[0]^2 + a[1]^2 + a[2]^2)						   //
//																   //
//	INPUT: Passed as parameters:								   //
//																   //
//		*Vector			float	Pointer to vector data			   //
//		*Len:			float	Pointer to store length			   //
//																   //
//	OUTPUT:	Return value is errno, math error code defined in	   //
//			C standard library.									   //
//																   //
//			Returns 0 if no math error.							   //
//																   //
/////////////////////////////////////////////////////////////////////

int VectorLength3d(
				   float	*Vector,
				   float	*Len)
{
	//local declarations
	float	t1;
	float	t2;
	float	t3;

	t1 = *Vector * *Vector;
	t2 = *(Vector + 1) * *(Vector + 1);
	t3 = *(Vector + 2) * *(Vector + 2);

	*Len = (float)sqrt(t1 + t2 + t3);

	return(0);

} // End VectorLength3d

//////////////////////// VectorAngle2d //////////////////////////////
//																   //
//	This function computes the angle between two 2d vectors.	   //
//	Actually, to save cycles, the cosine of the angle is returned  //
//	so if the user actually needs the angle, the acos should be	   //
//	computed by the calling function.							   //
//																   //
//	Also, this function uses several dot products.  Rather than	   //
//	calling the dot routine in this library, the dot products	   //
//	are computed explicitly.									   //
//																   //
//	Cos(angle) = A dot B / [ sqrt(A dot A) sqrt(B dot B) ]	       //
//																   //
//	INPUT: Passed as parameters:								   //
//																   //
//			*VectorA	float	Pointer to vector A data		   //
//			*VectorB	float	Pointer to vector B data		   //
//			*CosAngle	double	Pointer to resulting cos(angle)	   //
//																   //
//	OUTPUT:	Return value is errno, math error code defined in	   //
//			C standard library.									   //
//																   //
//			Returns 0 if no math error.							   //
//																   //
/////////////////////////////////////////////////////////////////////

int VectorAngle2d(
				   float	*VectorA,
				   float	*VectorB,
				   double	*CosAngle)
{

	//local declarations
	//doubles
	double DotA;				// A dot A
	double DotB;				// B dot B
	double DotAB;				// A dot B

	//zero the dot products
	DotA  = 0;
	DotB  = 0;
	DotAB = 0;

	//compute the dot products
	DotA  += (double)(*VectorA * *VectorA + 
				*(VectorA + 1) * *(VectorA + 1));
	DotAB += (double)(*VectorA * *VectorB + 
				*(VectorA + 1) * *(VectorB + 1));
	DotB  += (double)(*VectorB * *VectorB + 
				*(VectorB + 1) * *(VectorB + 1));

	

	//compute the bottom part of the equation
	//[bottom part] = sqrt(A dot A) * sqrt(B dot B)
	DotA  = sqrt(DotA);
	DotB  = sqrt(DotB);
	DotB *= DotA;

	//compute the cosine of the angle as A dot B / [ bottom part ]
	*CosAngle = DotAB/DotB;

	//return any math error
	return(errno);

}  // End VectorAngle2d

//////////////////////// VectorAngle3d //////////////////////////////
//																   //
//	This function computes the angle between two 3d vectors.	   //
//	Actually, to save cycles, the cosine of the angle is returned  //
//	so if the user actually needs the angle, the acos should be	   //
//	computed by the calling function.							   //
//																   //
//	Also, this function uses several dot products.  Rather than	   //
//	calling the dot routine in this library, the dot products	   //
//	are computed explicitly.									   //
//																   //
//	Cos(angle) = A dot B / [ sqrt(A dot A) sqrt(B dot B) ]	       //
//																   //
//	INPUT: Passed as parameters:								   //
//																   //
//			*VectorA	float	Pointer to vector A data		   //
//			*VectorB	float	Pointer to vector B data		   //
//			*CosAngle	double	Pointer to resulting cos(angle)	   //
//																   //
//	OUTPUT:	Return value is errno, math error code defined in	   //
//			C standard library.									   //
//																   //
//			Returns 0 if no math error.							   //
//																   //
/////////////////////////////////////////////////////////////////////

int VectorAngle3d(
				   float	*VectorA,
				   float	*VectorB,
				   double	*CosAngle)
{

	//local declarations
	//doubles
	double DotA;				// A dot A
	double DotB;				// B dot B
	double DotAB;				// A dot B

	//zero the dot products
	DotA  = 0;
	DotB  = 0;
	DotAB = 0;

	//compute the dot products
	DotA  += (double)(*VectorA * *VectorA + 
				*(VectorA + 1) * *(VectorA + 1) + 
				*(VectorA + 2) * *(VectorA + 2));
	DotAB += (double)(*VectorA * *VectorB + 
				*(VectorA + 1) * *(VectorB + 1) +
				*(VectorA + 2) * *(VectorB + 2));
	DotB  += (double)(*VectorB * *VectorB + 
				*(VectorB + 1) * *(VectorB + 1) +
				*(VectorB + 2) * *(VectorB + 2));

	

	//compute the bottom part of the equation
	//[bottom part] = sqrt(A dot A) * sqrt(B dot B)
	DotA  = sqrt(DotA);
	DotB  = sqrt(DotB);
	DotB *= DotA;

	//compute the cosine of the angle as A dot B / [ bottom part ]
	*CosAngle = DotAB/DotB;

	//return any math error
	return(errno);

}  // End VectorAngle3d
