///////////////////// calculus Implementation //////////////////////
//                                                                 //
//                   x87 based Matrix Routines                     //
//                                                                 //
//  Written by:  John S. Riley                                     //
//                                                                 //
//  This file contains a variety of differentiation and			   //
//	integration routines written for generic x87 floating-point	   //
//	processors.											           //
//                                                                 //
//  The routines use no explicit SIMD instructions, but may be     //
//  compiled using compiler generated SSE or 3dNow instructions.   //
//                                                                 //
//  The routines are general, using rolled up, variable		       //
//  sized loops.  Dimension and data are passed to the		       //
//  functions as parameters (data or pointers).                    //
//                                                                 //
/////////////////////////////////////////////////////////////////////

///////////////////////// Functions in this File ////////////////////
//                                                                 //
//  genTrap:			generic Trapezoidal Rule Integration       //
//	Trap100:			Trapezoidal Rule integration for n=100     //
//	Trap500:			Trapezoidal Rule integration for n=500     //
//	Trap512:			Trapezoidal Rule integration for n=512     //
//	Trap1000:			Trapezoidal Rule integration for n=1000    //
//	Trap1024:			Trapezoidal Rule integration for n=1024    //
//	Trap2048:			Trapezoidal Rule integration for n=2048    //
//	Trap4096:			Trapezoidal Rule integration for n=4096    //
//	Trap5000:			Trapezoidal Rule integration for n=5000    //  
//                                                                 //
/////////////////////////////////////////////////////////////////////

///////////////////////// INCLUDES //////////////////////////////////

#include "calculus.h"
#include "stdlib.h"
#include "math.h"

///////////////////////// FUNCTIONS /////////////////////////////////

/////////////////////// genTrap /////////////////////////////////////
//																   //
//	This function integrates a function from a to b using		   //
//	a generic Trapezoidal Rule routine unrolled by 16 terms.	   //
//																   //
//	I = h(1/2 f(0) + f(1) + f(2) + ... + f(n-1) + 1/2 f(n)		   //
//																   //
//	where h = (b-a)/n and the explicit function contains n terms   //
//																   //
//	NOTE:	Since this function uses a loop unrolled by 16 terms,  //
//			n MUST be greater than 18!!							   //
//																   //
//	INPUT: Passed as parameters:								   //
//																   //
//			*fx			float	function data					   //
//			a			float	start of integration interval	   //
//			b			float	end of integration interavl		   //
//			n			int		number of terms in function		   //
//			*I			float	Result							   //
//																   //
//	OUTPUT:	Return value is errno, math error code defined in	   //
//			C standard library.									   //
//																   //
//			Returns 0 if no math error.							   //
//																   //
/////////////////////////////////////////////////////////////////////

int genTrap(
			float	*fx,
			float	a,
			float	b,
			int		n,
			float	*I)
{

	//local declarations
	//integers
	int		j;					//summation counter
	int		loopbreak;			//max number of unrolled terms
	int		loopend;			//end of loop to maximize unrolling

	float	h;					//interval/#subintervals
	float	s1;					//sum of first and last terms
	float	s2;					//sum of inner terms

	//compute h
	h = (b-a)/n;

	//for pointer math, we really need n-1
	n -= 1;

	//add first and last terms, and divide by 2
	s1 = *fx + *(fx + n);
	s1 /= 2;

	//zero s2 since it is a running sum
	s2 = 0;

	//compute where the loop should end
	loopbreak = ((int)(n/16) * 16) + 1;
	loopend = loopbreak - 16;

	//compute the inner sum
	for(j = 1; j <= loopend; j += 16){

		s2 += *(fx + j) + *(fx + j + 1) + *(fx + j + 2) + *(fx + j + 3)
			+ *(fx + j + 4) + *(fx + j + 5) + *(fx + j + 6) + *(fx + j + 7)
			+ *(fx + j + 8) + *(fx + j + 9) + *(fx + j + 10) + *(fx + j + 11)
			+ *(fx + j + 12) + *(fx + j + 13) + *(fx + j + 14)
			+ *(fx + j + 15);

	}

	//clean-up terms that did not fit in main unrolled loop
	for(j = loopend; j <= n-1; ++j){

		s2 += *(fx + j);

	}

	//compute the integral from the summed terms
	*I = s1 + s2;
	*I *= h;

	//return any math error
	return(errno);

}  // end genTrap

/////////////////////// Trap100 /////////////////////////////////////
//																   //
//	This function integrates a function with 100 elements from a   //
//	to b using a Trapezoidal Rule routine unrolled by 16 terms.	   //
//																   //
//	I = h(1/2 f(0) + f(1) + f(2) + ... + f(n-1) + 1/2 f(n)		   //
//																   //
//	where h = (b-a)/100											   //
//																   //
//	INPUT: Passed as parameters:								   //
//																   //
//			*fx			float	function data					   //
//			a			float	start of integration interval	   //
//			b			float	end of integration interavl		   //
//			*I			float	Result							   //
//																   //
//	OUTPUT:	Return value is errno, math error code defined in	   //
//			C standard library.									   //
//																   //
//			Returns 0 if no math error.							   //
//																   //
/////////////////////////////////////////////////////////////////////

int Trap100(
			float	*fx,
			float	a,
			float	b,
			float	*I)
{

	//local declarations
	//integers
	int		j;

	//floats
	float	h;
	float	s1;
	float	s2;

	//compute h
	h = (b-a)/100;

	//add first and last terms, and divide by 2
	s1 = *fx + *(fx + 99);
	s1 /= 2;

	//zero s2 since it is a running sum
	s2 = 0;

	//compute the inner sum
	for(j = 1; j <= 81; j += 16){

		s2 += *(fx + j) + *(fx + j + 1) + *(fx + j + 2) + *(fx + j + 3)
			+ *(fx + j + 4) + *(fx + j + 5) + *(fx + j + 6) + *(fx + j + 7)
			+ *(fx + j + 8) + *(fx + j + 9) + *(fx + j + 10) + *(fx + j + 11)
			+ *(fx + j + 12) + *(fx + j + 13) + *(fx + j + 14)
			+ *(fx + j + 15);

	}

	//clean-up terms not in unrolled loop
	s2 += *(fx + 97) + *(fx + 98);

	//compute the integral from the summed terms
	*I = s1 + s2;
	*I *= h;

	//return any math error
	return(errno);

}  // end Trap100

/////////////////////// Trap512 /////////////////////////////////////
//																   //
//	This function integrates a function with 512 elements from a   //
//	to b using a Trapezoidal Rule routine unrolled by 16 terms.	   //
//																   //
//	I = h(1/2 f(0) + f(1) + f(2) + ... + f(n-1) + 1/2 f(n)		   //
//																   //
//	where h = (b-a)/512											   //
//																   //
//	INPUT: Passed as parameters:								   //
//																   //
//			*fx			float	function data					   //
//			a			float	start of integration interval	   //
//			b			float	end of integration interavl		   //
//			*I			float	Result							   //
//																   //
//	OUTPUT:	Return value is errno, math error code defined in	   //
//			C standard library.									   //
//																   //
//			Returns 0 if no math error.							   //
//																   //
/////////////////////////////////////////////////////////////////////

int Trap512(
			float	*fx,
			float	a,
			float	b,
			float	*I)
{

	//local declarations
	//integers
	int		j;

	//floats
	float	h;
	float	s1;
	float	s2;

	//compute h
	h = (b-a)/512;

	//add first and last terms, and divide by 2
	s1 = *fx + *(fx + 511);
	s1 /= 2;

	//zero s2 since it is a running sum
	s2 = 0;

	//compute the inner sum
	for(j = 1; j <= 481; j += 16){

		s2 += *(fx + j) + *(fx + j + 1) + *(fx + j + 2) + *(fx + j + 3)
			+ *(fx + j + 4) + *(fx + j + 5) + *(fx + j + 6) + *(fx + j + 7)
			+ *(fx + j + 8) + *(fx + j + 9) + *(fx + j + 10) + *(fx + j + 11)
			+ *(fx + j + 12) + *(fx + j + 13) + *(fx + j + 14)
			+ *(fx + j + 15);

	}

	//clean-up terms not in unrolled loop
	s2 += *(fx + 497) + *(fx + 498) + *(fx + 499) + *(fx + 500) + *(fx + 501)
		+ *(fx + 502) + *(fx + 503) + *(fx + 504) + *(fx + 505) 
		+ *(fx + 506) + *(fx + 507) + *(fx + 508) + *(fx + 509) 
		+ *(fx + 510);

	//compute the integral from the summed terms
	*I = s1 + s2;
	*I *= h;

	//return any math error
	return(errno);

}  // end Trap512

/////////////////////// Trap500 /////////////////////////////////////
//																   //
//	This function integrates a function with 500 elements from a   //
//	to b using a Trapezoidal Rule routine unrolled by 16 terms.	   //
//																   //
//	I = h(1/2 f(0) + f(1) + f(2) + ... + f(n-1) + 1/2 f(n)		   //
//																   //
//	where h = (b-a)/500											   //
//																   //
//	INPUT: Passed as parameters:								   //
//																   //
//			*fx			float	function data					   //
//			a			float	start of integration interval	   //
//			b			float	end of integration interavl		   //
//			*I			float	Result							   //
//																   //
//	OUTPUT:	Return value is errno, math error code defined in	   //
//			C standard library.									   //
//																   //
//			Returns 0 if no math error.							   //
//																   //
/////////////////////////////////////////////////////////////////////

int Trap500(
			float	*fx,
			float	a,
			float	b,
			float	*I)
{

	//local declarations
	//integers
	int		j;

	//floats
	float	h;
	float	s1;
	float	s2;

	//compute h
	h = (b-a)/500;

	//add first and last terms, and divide by 2
	s1 = *fx + *(fx + 499);
	s1 /= 2;

	//zero s2 since it is a running sum
	s2 = 0;

	//compute the inner sum
	for(j = 1; j <= 481; j += 16){

		s2 += *(fx + j) + *(fx + j + 1) + *(fx + j + 2) + *(fx + j + 3)
			+ *(fx + j + 4) + *(fx + j + 5) + *(fx + j + 6) + *(fx + j + 7)
			+ *(fx + j + 8) + *(fx + j + 9) + *(fx + j + 10) + *(fx + j + 11)
			+ *(fx + j + 12) + *(fx + j + 13) + *(fx + j + 14)
			+ *(fx + j + 15);

	}

	//clean-up terms not in unrolled loop
	s2 += *(fx + 497) + *(fx + 498);

	//compute the integral from the summed terms
	*I = s1 + s2;
	*I *= h;

	//return any math error
	return(errno);

}  // end Trap500

/////////////////////// Trap1000 ////////////////////////////////////
//																   //
//	This function integrates a function with 1000 elements from a  //
//	to b using a Trapezoidal Rule routine unrolled by 16 terms.	   //
//																   //
//	I = h(1/2 f(0) + f(1) + f(2) + ... + f(n-1) + 1/2 f(n)		   //
//																   //
//	where h = (b-a)/1000										   //
//																   //
//	INPUT: Passed as parameters:								   //
//																   //
//			*fx			float	function data					   //
//			a			float	start of integration interval	   //
//			b			float	end of integration interavl		   //
//			*I			float	Result							   //
//																   //
//	OUTPUT:	Return value is errno, math error code defined in	   //
//			C standard library.									   //
//																   //
//			Returns 0 if no math error.							   //
//																   //
/////////////////////////////////////////////////////////////////////

int Trap1000(
			float	*fx,
			float	a,
			float	b,
			float	*I)
{

	//local declarations
	//integers
	int		j;

	//floats
	float	h;
	float	s1;
	float	s2;

	//compute h
	h = (b-a)/1000;

	//add first and last terms, and divide by 2
	s1 = *fx + *(fx + 999);
	s1 /= 2;

	//zero s2 since it is a running sum
	s2 = 0;

	//compute the inner sum
	for(j = 1; j <= 977; j += 16){

		s2 += *(fx + j) + *(fx + j + 1) + *(fx + j + 2) + *(fx + j + 3)
			+ *(fx + j + 4) + *(fx + j + 5) + *(fx + j + 6) + *(fx + j + 7)
			+ *(fx + j + 8) + *(fx + j + 9) + *(fx + j + 10) + *(fx + j + 11)
			+ *(fx + j + 12) + *(fx + j + 13) + *(fx + j + 14)
			+ *(fx + j + 15);

	}

	//clean-up terms not in unrolled loop
	s2 += *(fx + 993) + *(fx + 994) + *(fx + 995) + *(fx + 996) + *(fx + 997)
		+ *(fx + 998);

	//compute the integral from the summed terms
	*I = s1 + s2;
	*I *= h;

	//return any math error
	return(errno);

}  // end Trap1000

/////////////////////// Trap1024 ////////////////////////////////////
//																   //
//	This function integrates a function with 1024 elements from a  //
//	to b using a Trapezoidal Rule routine unrolled by 16 terms.	   //
//																   //
//	I = h(1/2 f(0) + f(1) + f(2) + ... + f(n-1) + 1/2 f(n)		   //
//																   //
//	where h = (b-a)/1024										   //
//																   //
//	INPUT: Passed as parameters:								   //
//																   //
//			*fx			float	function data					   //
//			a			float	start of integration interval	   //
//			b			float	end of integration interavl		   //
//			*I			float	Result							   //
//																   //
//	OUTPUT:	Return value is errno, math error code defined in	   //
//			C standard library.									   //
//																   //
//			Returns 0 if no math error.							   //
//																   //
/////////////////////////////////////////////////////////////////////

int Trap1024(
			float	*fx,
			float	a,
			float	b,
			float	*I)
{

	//local declarations
	//integers
	int		j;

	//floats
	float	h;
	float	s1;
	float	s2;

	//compute h
	h = (b-a)/1024;

	//add first and last terms, and divide by 2
	s1 = *fx + *(fx + 1023);
	s1 /= 2;

	//zero s2 since it is a running sum
	s2 = 0;

	//compute the inner sum
	for(j = 1; j <= 993; j += 16){

		s2 += *(fx + j) + *(fx + j + 1) + *(fx + j + 2) + *(fx + j + 3)
			+ *(fx + j + 4) + *(fx + j + 5) + *(fx + j + 6) + *(fx + j + 7)
			+ *(fx + j + 8) + *(fx + j + 9) + *(fx + j + 10) + *(fx + j + 11)
			+ *(fx + j + 12) + *(fx + j + 13) + *(fx + j + 14)
			+ *(fx + j + 15);

	}

	//clean-up terms not in unrolled loop
	s2 += *(fx + 1009) + *(fx + 1010) + *(fx + 1011) + *(fx + 1012) + *(fx + 1013)
		+ *(fx + 1014) + *(fx + 1015) + *(fx + 1016) + *(fx + 1017) + *(fx + 1018) 
		+ *(fx + 1019) + *(fx + 1020) + *(fx + 1021) + *(fx + 1022);

	//compute the integral from the summed terms
	*I = s1 + s2;
	*I *= h;

	//return any math error
	return(errno);

}  // end Trap1024

/////////////////////// Trap2048 ////////////////////////////////////
//																   //
//	This function integrates a function with 2048 elements from a  //
//	to b using a Trapezoidal Rule routine unrolled by 16 terms.	   //
//																   //
//	I = h(1/2 f(0) + f(1) + f(2) + ... + f(n-1) + 1/2 f(n)		   //
//																   //
//	where h = (b-a)/2048										   //
//																   //
//	INPUT: Passed as parameters:								   //
//																   //
//			*fx			float	function data					   //
//			a			float	start of integration interval	   //
//			b			float	end of integration interavl		   //
//			*I			float	Result							   //
//																   //
//	OUTPUT:	Return value is errno, math error code defined in	   //
//			C standard library.									   //
//																   //
//			Returns 0 if no math error.							   //
//																   //
/////////////////////////////////////////////////////////////////////

int Trap2048(
			float	*fx,
			float	a,
			float	b,
			float	*I)
{

	//local declarations
	//integers
	int		j;

	//floats
	float	h;
	float	s1;
	float	s2;

	//compute h
	h = (b-a)/2048;

	//add first and last terms, and divide by 2
	s1 = *fx + *(fx + 2047);
	s1 /= 2;

	//zero s2 since it is a running sum
	s2 = 0;

	//compute the inner sum
	for(j = 1; j <= 2017; j += 16){

		s2 += *(fx + j) + *(fx + j + 1) + *(fx + j + 2) + *(fx + j + 3)
			+ *(fx + j + 4) + *(fx + j + 5) + *(fx + j + 6) + *(fx + j + 7)
			+ *(fx + j + 8) + *(fx + j + 9) + *(fx + j + 10) + *(fx + j + 11)
			+ *(fx + j + 12) + *(fx + j + 13) + *(fx + j + 14)
			+ *(fx + j + 15);

	}

	//clean-up terms not in unrolled loop
	s2 += *(fx + 2033) + *(fx + 2034) + *(fx + 2035) + *(fx + 2036) + *(fx + 2037)
		+ *(fx + 2038) + *(fx + 2039) + *(fx + 2040) + *(fx + 2041) + *(fx + 2042) 
		+ *(fx + 2043) + *(fx + 2044) + *(fx + 2045) + *(fx + 2046);

	//compute the integral from the summed terms
	*I = s1 + s2;
	*I *= h;

	//return any math error
	return(errno);

}  // end Trap2048

/////////////////////// Trap4096 ////////////////////////////////////
//																   //
//	This function integrates a function with 4096 elements from a  //
//	to b using a Trapezoidal Rule routine unrolled by 16 terms.	   //
//																   //
//	I = h(1/2 f(0) + f(1) + f(2) + ... + f(n-1) + 1/2 f(n)		   //
//																   //
//	where h = (b-a)/4096										   //
//																   //
//	INPUT: Passed as parameters:								   //
//																   //
//			*fx			float	function data					   //
//			a			float	start of integration interval	   //
//			b			float	end of integration interavl		   //
//			*I			float	Result							   //
//																   //
//	OUTPUT:	Return value is errno, math error code defined in	   //
//			C standard library.									   //
//																   //
//			Returns 0 if no math error.							   //
//																   //
/////////////////////////////////////////////////////////////////////

int Trap4096(
			float	*fx,
			float	a,
			float	b,
			float	*I)
{

	//local declarations
	//integers
	int		j;

	//floats
	float	h;
	float	s1;
	float	s2;

	//compute h
	h = (b-a)/4096;

	//add first and last terms, and divide by 2
	s1 = *fx + *(fx + 4095);
	s1 /= 2;

	//zero s2 since it is a running sum
	s2 = 0;

	//compute the inner sum
	for(j = 1; j <= 4065; j += 16){

		s2 += *(fx + j) + *(fx + j + 1) + *(fx + j + 2) + *(fx + j + 3)
			+ *(fx + j + 4) + *(fx + j + 5) + *(fx + j + 6) + *(fx + j + 7)
			+ *(fx + j + 8) + *(fx + j + 9) + *(fx + j + 10) + *(fx + j + 11)
			+ *(fx + j + 12) + *(fx + j + 13) + *(fx + j + 14)
			+ *(fx + j + 15);

	}

	//clean-up terms not in unrolled loop
	s2 += *(fx + 4081) + *(fx + 4082) + *(fx + 4083) + *(fx + 4084) + *(fx + 4085)
		+ *(fx + 4086) + *(fx + 4087) + *(fx + 4088) + *(fx + 4089) + *(fx + 4090) 
		+ *(fx + 4091) + *(fx + 4092) + *(fx + 4093) + *(fx + 4094);

	//compute the integral from the summed terms
	*I = s1 + s2;
	*I *= h;

	//return any math error
	return(errno);

}  // end Trap4096

/////////////////////// Trap5000 ////////////////////////////////////
//																   //
//	This function integrates a function with 5000 elements from a  //
//	to b using a Trapezoidal Rule routine unrolled by 16 terms.	   //
//																   //
//	I = h(1/2 f(0) + f(1) + f(2) + ... + f(n-1) + 1/2 f(n)		   //
//																   //
//	where h = (b-a)/5000										   //
//																   //
//	INPUT: Passed as parameters:								   //
//																   //
//			*fx			float	function data					   //
//			a			float	start of integration interval	   //
//			b			float	end of integration interavl		   //
//			*I			float	Result							   //
//																   //
//	OUTPUT:	Return value is errno, math error code defined in	   //
//			C standard library.									   //
//																   //
//			Returns 0 if no math error.							   //
//																   //
/////////////////////////////////////////////////////////////////////

int Trap5000(
			float	*fx,
			float	a,
			float	b,
			float	*I)
{

	//local declarations
	//integers
	int		j;

	//floats
	float	h;
	float	s1;
	float	s2;

	//compute h
	h = (b-a)/5000;

	//add first and last terms, and divide by 2
	s1 = *fx + *(fx + 4999);
	s1 /= 2;

	//zero s2 since it is a running sum
	s2 = 0;

	//compute the inner sum
	for(j = 1; j <= 4977; j += 16){

		s2 += *(fx + j) + *(fx + j + 1) + *(fx + j + 2) + *(fx + j + 3)
			+ *(fx + j + 4) + *(fx + j + 5) + *(fx + j + 6) + *(fx + j + 7)
			+ *(fx + j + 8) + *(fx + j + 9) + *(fx + j + 10) + *(fx + j + 11)
			+ *(fx + j + 12) + *(fx + j + 13) + *(fx + j + 14)
			+ *(fx + j + 15);

	}

	//clean-up terms not in unrolled loop
	s2 += *(fx + 4993) + *(fx + 4994) + *(fx + 4995) + *(fx + 4996) + *(fx + 4997)
		+ *(fx + 4998);

	//compute the integral from the summed terms
	*I = s1 + s2;
	*I *= h;

	//return any math error
	return(errno);

}  // end Trap5000