///////////////////// genVector Implementation //////////////////////
//                                                                 //
//                   x87 based Vector Routines                     //
//                                                                 //
//  Written by:  John S. Riley                                     //
//                                                                 //
//  This file contains a variety of vector manipulation routines   //
//  written for generic x87 floating-point processors.             //
//                                                                 //
//  The routines use no explicit SIMD instructions, but may be     //
//  compiled using compiler generated SSE or 3dNow instructions.   //
//                                                                 //
//  The vector routines are general, using rolled up, variable     //
//  sized loops. Vector dimension and data are passed to the       //
//  functions as parameters (data or pointers).                    //
//                                                                 //
/////////////////////////////////////////////////////////////////////

///////////////////////// Functions in this File ////////////////////
//                                                                 //
//  genVectorDotProd:   Vector Dot Product                         //
//	genVectorLength:	Length of Vector						   //
//	genVectorCrossProd: Vector Cross Product					   //
//	genVectorRotate:	Rotate vector around x,y,z axes			   //
//	genVectorTranslate:	Translate vector; change axis origin	   //
//	genVectorScale:	    Multiply vector by a scalar				   //
//	genVectorAdd:		Add two vectors							   //
//	genVectorAngle:		Angle between two vectors				   //
//  genVectorCrossProdLen:	length of cross product vector		   //
//	genVectorNegate:	Negate of Vector						   //  
//                                                                 //
/////////////////////////////////////////////////////////////////////

///////////////////////// INCLUDES //////////////////////////////////

#include "genVector.h"
#include "stdlib.h"
#include "math.h"

///////////////////////// FUNCTIONS /////////////////////////////////

//////////////////////// genVectorDotProd ///////////////////////////
//																   //
//	This function computes the dot product of two generic 		   //
//  vectors														   //
//																   //
//	C = A . B, computed as C = sum{a[j]b[j]}					   //
//																   //
//	INPUT: Passed as parameters:								   //
//																   //
//			*VectorA	float	Pointer to A data				   //
//			*VectorB	float	Pointer to B data				   //
//			dimen		int		dimension vectors A and B		   //
//			*Result		float	Pointer to result, C			   //
//																   //
//	OUTPUT:	Return value is errno, math error code defined in	   //
//			C standard library.									   //
//																   //
//			Returns 0 if no math error.							   //
//																   //
/////////////////////////////////////////////////////////////////////

int genVectorDotProd(
					 float	*VectorA,
					 float	*VectorB,
					 int	dimen,
					 float	*Result)
{

	//local declarations
	//integers
	int j;				//counter for vector components

	//zero the result
	*Result = 0;

	//loop over vector components
	for(j=0; j < dimen; ++j){

		//build the dot product as sum{a[j]b[j]}
		*Result += *(VectorA + j) * *(VectorB + j);
		
	}

	//return any math error
	return(errno);

} // End genVectorDotProd

//////////////////////// genVectorLength ////////////////////////////
//																   //
//	This function finds the length of a generic vector			   //
//																   //
//	length = sqrt(sum{a[j]^2})									   //
//																   //
//	INPUT: Passed as parameters:								   //
//																   //
//			*Vector		float	Pointer data					   //
//			dimen		int		dimension vectors A				   //
//			*Length		float	Pointer to resulting length		   //
//																   //
//	OUTPUT:	Return value is errno, math error code defined in	   //
//			C standard library.									   //
//																   //
//			Returns 0 if no math error.							   //
//																   //
/////////////////////////////////////////////////////////////////////

int genVectorLength(
					float	*Vector,
					int		dimen,
					float	*Length)
{

	//local declarations
	//integers
	int j;				// vector component counter
	
	//zero the length
	*Length = 0;

	//loop over vector components
	for(j=0; j < dimen; ++j){

		//build the radicand, sum{a[j]^2}
		*Length += *(Vector + j) * *(Vector + j);

	}

	//compute the actual length
	*Length = (float)sqrt(*Length);

	//return any math error
	return(errno);

} // End genVectorLength

//////////////////////// genVectorScale /////////////////////////////
//																   //
//	This function multiplies a generic vector by a scalar		   //
//																   //
//	C = sA, where c[1] = sa[1], c[2] = sa[2], ...				   //
//																   //
//	INPUT: Passed as parameters:								   //
//																   //
//			*Vector		float	Pointer to vector data			   //
//			dimen		int		dimension of vector				   //
//			scalar		float	scalar to multiply by vector	   //
//			*Result		float	Pointer to resulting scaled vector //
//																   //
//	OUTPUT:	Return value is errno, math error code defined in	   //
//			C standard library.									   //
//																   //
//			Returns 0 if no math error.							   //
//																   //
/////////////////////////////////////////////////////////////////////

int genVectorScale(
				   float	*Vector,
				   int		dimen,
				   float	scalar,
				   float	*ScaledVector)
{

	//local declarations
	//integers
	int	j;				//counter for vector components

	//loop over vector components
	for(j = 0; j < dimen; ++j){

		//compute c[j] = a[j] * s
		*(ScaledVector + j) = *(Vector + j) * scalar;

	}

	//return any math error
	return(errno);

} // End genVectorScale

//////////////////////// genVectorAdd ///////////////////////////////
//																   //
//	This function adds two generic vectors						   //
//																   //
//	C = A + B, where c[1] = a[1] + b[1], c[2] = a[2] + b[2], ...   //
//																   //
//	INPUT: Passed as parameters:								   //
//																   //
//			*VectorA	float	Pointer to A data				   //
//			*VectorB	float	Pointer to B data				   //
//			dimen		int		dimension of vectors A	and B	   //
//			*VectorSum	float	Pointer to resultant vector		   //
//																   //
//	OUTPUT:	Return value is errno, math error code defined in	   //
//			C standard library.									   //
//																   //
//			Returns 0 if no math error.							   //
//																   //
/////////////////////////////////////////////////////////////////////

int genVectorAdd(
				 float	*VectorA,
				 float	*VectorB,
				 int	dimen,
				 float	*VectorSum)
{

	//local declarations
	//integers
	int j;				//vector component counter

	//loop over vector components
	for(j=0; j < dimen; ++j){

		//compute sum, c[j] = a[j] + b[j]
		*(VectorSum + j) = *(VectorA + j) + *(VectorB + j);

	}

	//return any math error
	return(errno);

} // End genVectorAdd

//////////////////////// genVectorAngle /////////////////////////////
//																   //
//	This function computes the angle between two vectors.		   //
//	Actually, to save cycles, the cosine of the angle is returned  //
//	so if the user actually needs the angle, the acos should be	   //
//	computed by the calling function.							   //
//																   //
//	Also, this function uses several dot products.  Rather than	   //
//	calling the dot routine in this library, the dot products	   //
//	are computed explicitly.									   //
//																   //
//	Cos(angle) = A dot B / [ sqrt(A dot A) sqrt(B dot B) ]	       //
//																   //
//	INPUT: Passed as parameters:								   //
//																   //
//			*VectorA	float	Pointer to A data				   //
//			*VectorB	float	Pointer to B data				   //
//			dimen		int		dimension of vectors A	and B	   //
//			*CosAngle	double	Pointer to resulting cos(angle)	   //
//																   //
//	OUTPUT:	Return value is errno, math error code defined in	   //
//			C standard library.									   //
//																   //
//			Returns 0 if no math error.							   //
//																   //
/////////////////////////////////////////////////////////////////////

int genVectorAngle(
				   float	*VectorA,
				   float	*VectorB,
				   int		dimen,
				   double	*CosAngle)
{

	//local declarations
	//integers
	int j;					//vector component counter

	//doubles
	double DotA;				// A dot A
	double DotB;				// B dot B
	double DotAB;			// A dot B

	//zero the dot products
	DotA  = 0;
	DotB  = 0;
	DotAB = 0;

	//loop over vector components
	for(j=0; j < dimen; ++j){

		//compute the dot products
		DotA  += (double)(*(VectorA + j) * *(VectorA + j));
		DotAB += (double)(*(VectorA + j) * *(VectorB + j));
		DotB  += (double)(*(VectorB + j) * *(VectorB + j));

	}

	//compute the bottom part of the equation
	//[bottom part] = sqrt(A dot A) * sqrt(B dot B)
	DotA  = sqrt(DotA);
	DotB  = sqrt(DotB);
	DotB *= DotA;

	//compute the cosine of the angle as A dot B / [ bottom part ]
	*CosAngle = DotAB/DotB;

	//return any math error
	return(errno);

}  // End genVectorAngle

//////////////////////// genVectorCrossProdLen //////////////////////
//																   //
//	This function computes the length of the cross product		   //
//	between two vectors.										   //								   //
//																   //
//	Also, this function uses several dot products.  Rather than	   //
//	calling the dot routine in this library, the dot products	   //
//	are computed explicitly.									   //
//																   //
//	Length = sqrt{ (A dot A)(B dot B) + (A dot B)^2 }   	       //
//																   //
//	INPUT: Passed as parameters:								   //
//																   //
//			*VectorA	float	Pointer to A data				   //
//			*VectorB	float	Pointer to B data				   //
//			dimen		int		dimension of vectors A	and B	   //
//			*CrossProdLen float	Pointer to Length of Cross Prod	   //
//																   //
//	OUTPUT:	Return value is errno, math error code defined in	   //
//			C standard library.									   //
//																   //
//			Returns 0 if no math error.							   //
//																   //
/////////////////////////////////////////////////////////////////////

int genVectorCrossProdLen(
						  float	*VectorA,
						  float	*VectorB,
						  int	dimen,
						  float	*CrossProdLen)
{

	//local declarations
	//integers
	int j;				//vector component counter

	//doubles
	float	DotA;		// A dot A
	float	DotB;		// B dot B
	float	DotAB;		// A dot B

	//zero the dot products
	DotA	= 0;
	DotB	= 0;
	DotAB	= 0;

	//loop over vector components
	for(j=0; j < dimen; ++j){

		//compute the three dot products
		DotA	+= *(VectorA + j) * *(VectorA + j);
		DotAB	+= *(VectorA + j) * *(VectorB + j);
		DotB	+= *(VectorB + j) * *(VectorB + j);

	}

	//compute the radicand, (A dot A)(B dot B) + (A dot B)^2
	DotAB	*=	DotAB;
	DotA	*=	DotB;
	DotA	+=	DotAB;

	//and take square root to get length
	*CrossProdLen = (float)sqrt((double)DotA);

	//return any math error
	return(errno);

} // End genVectorCrossProdLen

//////////////////////// genVectorNegate ///////////////////////////////
//																   //
//	This function computes the negative of a generic vector		   //
//																   //
//	B = -A														   //
//																   //
//	INPUT: Passed as parameters:								   //
//																   //
//			*VectorA	float	Pointer to A data				   //
//			*VectorB	float	Pointer to B data, -A				   //
//			dimen		int		dimension of vectors A	and B	   //
//																   //
//	OUTPUT:	Return value is errno, math error code defined in	   //
//			C standard library.									   //
//																   //
//			Returns 0 if no math error.							   //
//																   //
/////////////////////////////////////////////////////////////////////

int genVectorNegate(
				 float	*VectorA,
				 float	*VectorB,
				 int	dimen)
{

	//local declarations
	//integers
	int j;				//vector component counter

	//loop over vector components
	for(j=0; j < dimen; ++j){

		//compute negative, b[j] = -a[j]
		*(VectorB + j) = -1 * *(VectorA + j);

	}

	//return any math error
	return(errno);

} // End genVectorNegate

